# GPU BLAS Example

We will take the well known axpy vector operation from
[BLAS](https://netlib.org/blas):

```
    a * x + y
```

where `a` is a scalar and `x` and `y` are vectors.

This example both implements naive versions of saxpy and daxpy as well as uses
the tuned [cuBLAS](https://docs.nvidia.com/cuda/cublas/contents.html) provided
with CUDA.

## Code

The C++ CUDA code file `axpy.cu` defines the naive axpy kernels, allocates all
the arrays, and then benchmarks the naive implenentations vs. cuBLAS reporting
the average number of microseconds each took per call.

## Compiling

The code (`axpy.cu`) must be compiled ahead of time with the NVIDIA compiler
from the version of CUDA that will be used. Run `compile_nhr.sh` for the NHR
(Emmy/Grete) and `compile_scc.sh` for the SCC. Each will compile `axpy.cu`
usinig several different versions of the CUDA module available on the cluster
(this lets one compare performance between CUDA versions).

The compilation produces files of the form `axpy_CLUSTER_CUDAVERSION` where
`CLUSTER` is either `NHR` or `SCC` and `CUDAVERSION` is the CUDA version.

## Running

In the same directory that you ran compilation, submit the batch jobs with

``` shell
sbatch slurm_nhr.sh
```

on the NHR (Emmy/Grete) and

``` shell
sbatch slurm_scc.sh
```

on the SCC.

The SLURM job is a job array with one job per version of CUDA. Each one requests
a single node with a single GPU and CPU core. In the job scripts, the line:

``` bash
#SBATCH -a START-END
```

where `START` and `END` are non-decreasing integers tells SLURM to run a job
for each integer going from `START` to `END`, inclusive. In each job, the value
is passed as the environmental variable `SLURM_ARRAY_TASK_ID`. In these job
scripts, `SLURM_ARRAY_TASK_ID` is used by the logic to select the CUDA version
to use.

You can learn more about SLURM job arrays in the
[NHR documentation](https://www.hlrn.de/doc/display/PUB/Slurm#Slurm-JobArraysjobarray)
and [SCC documentation](https://docs.gwdg.de/doku.php?id=en:services:application_services:high_performance_computing:running_jobs_slurm:job_arrays).
